@@ -426,14 +426,16 @@ int check(CDFstatus status) {
 
 /* Helper function to handle logic of throwing exceptions when
  * allocations fail. */
-void *alloc(void *region) {
+void *
+alloc(void *region) {
     if (region == NULL) {
         PyErr_NoMemory();
     }
     return region;
 }
 
-void *rebinFromPythonToC(PyObject *in, long type) {
+void *
+rebinFromPythonToC(PyObject *in, long type) {
     if (in != NULL) {
         Py_IncRef(in);
         void *out = NULL;
@@ -457,6 +459,18 @@ void *rebinFromPythonToC(PyObject *in, long type) {
         } else {
             long len = PySequence_Check(in) ? PySequence_Size(in) : 1;
             long size = getSize(type);
+            if (size == -1) {
+                PyObject *item = in;
+                if (PySequence_Check(in)) {
+                    item = PySequence_GetItem(in, 0);
+                }
+                if (PyInt_Check(item) || PyLong_Check(item)) {
+                    type = CDF_INT4;
+                } else {
+                    type = CDF_FLOAT;
+                }
+                size = getSize(type);
+            }
             out = alloc(calloc(size, len));
 
             if (out != NULL) {
@@ -665,6 +679,10 @@ long getSize(long type) {
     return result;
 }
 
+long getNumber(int z) {
+    return longFromTwoTokens(GET_, (z ? zVAR_NUMELEMS_ : rVAR_NUMELEMS_));
+}
+
 
 
 /**
@@ -840,9 +858,26 @@ PyObject *tokenFormat_L_x(long one, long two, PyObject *tokens,
     return NULL;
 }
 
-PyObject *tokenFormat_p_x(long one, long two, PyObject *tokens,
+PyObject *tokenFormat_v_x(long one, long two, PyObject *tokens,
     long (*helper)(PyObject *)) {
-    return tokenFormat_l_x(one, two, tokens, helper);
+    if (helper != NULL) {
+        long type = helper(tokens);
+        if (type < 0 || type == CDF_CHAR) {
+            /* String handling has different ideas of "size". */
+            return tokenFormat_s_x(one, two, tokens, helper);
+        } else {
+            PyObject *in_1 = NULL;
+            if (PyArg_ParseTuple(tokens, "O", &in_1)) {
+                void **conv_1 = rebinFromPythonToC(in_1, type);
+                if (check(CDFlib(one, two, conv_1[0], NULL_))) {
+                    free(conv_1);
+                    return Py_None;
+                }
+                free(conv_1);
+            }
+        }
+    }
+    return NULL;
 }
 
 PyObject *tokenFormat_s_x(long one, long two, PyObject *tokens,
@@ -941,84 +976,165 @@ PyObject *tokenFormat_x_l(long one, long two, PyObject *tokens,
     return NULL;
 }
 
-PyObject *tokenCustom_zVAR_V(long one, long two, PyObject *tokens,
-    long (*helper)(PyObject *)) {
-    long i = 0;
+/* Hyper data functions */
+void **allocateHyperDataStorage(int z, long **dims, long *n_dims, long type) {
     // Assume the user has selected everything appropriately.  Determine
     // the size and configuration of the memory structure we need based
     // on the selections.
     // The number of dimensions is the same as the number of dimensions in
-    // the variable records plus a dimension if we are retrieving more
-    // than one record at a time.
-    long records = longFromTwoTokens(CONFIRM_, zVAR_RECCOUNT_);
-    long count = longFromTwoTokens(GET_, zVAR_NUMDIMS_);
-    long *dims = alloc(calloc(sizeof(long), count + 1));
+    // the variable records plus a dimension if we are setting or getting
+    // more than one record at a time.
+    long dimensionCount = *n_dims;
+    long records = longFromTwoTokens(
+      CONFIRM_, (z ? zVAR_RECCOUNT_ : rVARs_RECCOUNT_));
     long majority = longFromTwoTokens(GET_, CDF_MAJORITY_);
-    if (majority == ROW_MAJOR) {
-        dims[count] = records;
-    } else {
-        dims[0] = records;
+    long dimensionOffset = 0;
+    long recordDimension = ((majority == ROW_MAJOR) ? *n_dims : 0);
+    *n_dims += ((records > 1) ? 1 : 0);
+    long stringDimension = ((majority == ROW_MAJOR) ? 0 : *n_dims);
+    *n_dims += ((type == CDF_CHAR) ? 1 : 0);
+    *dims = alloc(calloc(sizeof(long), *n_dims));
+    long size = getSize(type);
+    if (records > 1) {
+        dims[recordDimension] = records;
+        dimensionOffset += (recordDimension == dimensionOffset);
     }
-    if (count > 0) {
-        long *lengths = longsFromTwoTokens(GET_, zVAR_DIMSIZES_);
-        for (i = 0; i < count; i++) {
+    if (type == CDF_CHAR) {
+        *dims[stringDimension] = 1;
+        size *= getNumber(z);
+        dimensionOffset += (stringDimension == dimensionOffset);
+    }
+    if (dimensionCount > 0) {
+        long *lengths = longsFromTwoTokens(
+          GET_, (z ? zVAR_DIMSIZES_ : rVARs_DIMSIZES_));
+        long i = 0;
+        for (i = 0; i < dimensionCount; i++) {
             if (majority == ROW_MAJOR) {
-                dims[count - i - 1] = lengths[i];
+                *dims[dimensionCount - i - 1 + dimensionOffset] = lengths[i];
             } else {
-                dims[i + 1] = lengths[i];
+                *dims[i + 1 + dimensionOffset] = lengths[i];
             }
         }
+        free(lengths);
     }
-    long type = typeHelper_zVAR_(NULL);
-    long size = getSize(type);
-    void **out_1 = multiDimensionalArray(dims, count + 1, size);
-    if (check(CDFlib(one, two, out_1, NULL_))) {
-        // Now we need to convert the buffer into the appropriate structure
-        // of arrays within arrays.
-        PyObject *conv_1 = NULL;
-        if (count == 0) {
-            conv_1 = ownedPythonListFromArray(NULL, 0, type);
-            PyList_Append(conv_1, castFromCdfToPython(type, out_1));
-        } else {
-            conv_1 = ownedPythonListOfListsFromArray(
-              out_1, dims, count + 1, type);
+    return multiDimensionalArray(*dims, *n_dims, size);
+}
+
+/* Get zVAR hyperdata */
+PyObject *getHyperData(int z, long one, long two) {
+    long n_dims = longFromTwoTokens(
+      GET_, (z ? zVAR_NUMDIMS_ : rVARs_NUMDIMS_));
+    long *dims = NULL;
+    long type = (z ? typeHelper_zVAR_(NULL) : typeHelper_rVAR_(NULL));
+    void **out_1 = allocateHyperDataStorage(z, &dims, &n_dims, type);
+    if (dims != NULL) {
+        if (out_1 != NULL) {
+            if (check(CDFlib(one, two, out_1, NULL_))) {
+                // Now we need to convert the buffer into the appropriate structure
+                // of arrays within arrays.
+                PyObject *conv_1 = NULL;
+                if (n_dims == 0) {
+                    conv_1 = ownedPythonListFromArray(NULL, 0, type);
+                    PyList_Append(conv_1, castFromCdfToPython(type, out_1));
+                } else {
+                    conv_1 = ownedPythonListOfListsFromArray(
+                      out_1, dims, n_dims, type);
+                }
+                cleanupMultiDimensionalArray(out_1, dims, n_dims);
+                free(dims);
+                return Py_BuildValue("(O)", conv_1);
+            }
+            cleanupMultiDimensionalArray(out_1, dims, n_dims);
         }
-        free(out_1);
-        return Py_BuildValue("(O)", conv_1);
+        free(dims);
     }
-    free(out_1);
     return NULL;
 }
 
-PyObject *tokenFormat_x_p(long one, long two, PyObject *tokens,
-    long (*helper)(PyObject *)) {
-    if (helper != NULL) {
-        long type = helper(tokens);
-        if (type < 0) {
-            return tokenFormat_x_s(one, two, tokens, helper);
-        } else {
-            long len = type % TYPE_MOD;
-            type -= len;
-            type /= TYPE_MOD;
-            if (len > 0) {
-                long size = getSize(type);
-                void *out_1 = alloc(calloc(size, len));
-                if (out_1 != NULL) {
-                    if (check(CDFlib(one, two, out_1, NULL_))) {
-                        /* Convert array into Python list. */
-                        PyObject *conv_1
-                          = ownedPythonListFromArray(out_1, len, type);
-                        free(out_1);
-                        return Py_BuildValue("(O)", conv_1);
-                    }
-                    free(out_1);
+void hyperDataFromOwnedPythonSequenceTrees(
+  void **buffer, PyObject *data, long *dims, long n_dims, long type) {
+    if ((n_dims <= 1) && (type != CDF_CHAR)) {
+        void **tmp = rebinFromPythonToC(data, type);
+        memcpy(buffer, tmp, getSize(type) * PyList_Size(data));
+        free(tmp);
+    } else if (n_dims > (type == CDF_CHAR ? 1 : 0)) {
+        long i = 0;
+        for (i = 0; i < PyList_Size(data); i++) {        
+            hyperDataFromOwnedPythonSequenceTrees(
+              buffer[i], PyList_GetItem(data, i), dims + 1, n_dims - 1, type);
+        }
+    } else {
+        if (PyList_Check(data) && (PyList_Size(data) == 1)) {
+            PyObject *item = PySequence_GetItem(data, 0);
+            /* Convert this from a sequence to a string before
+             * calling in to the rebin function, otherwise it
+             * will miss the string processing and get caught
+             * in the integer processing. */
+            if (PyString_Check(item) || PyUnicode_Check(item)) {
+                data = item;
+            }
+        }
+        if (data != NULL) {
+            void **tmp = rebinFromPythonToC(data, type);
+            if (tmp != NULL) {
+                long len;
+                if (type == CDF_CHAR) {
+                    len = strlen((const char *)tmp) + 1;
+                } else {
+                    len = getSize(type) * dims[0];
                 }
+                memcpy(buffer, tmp, len);
+                free(tmp);
             }
         }
     }
+}
+
+PyObject *setHyperData(int z, long one, long two, PyObject *tokens) {
+    PyObject *in_1 = NULL;
+    if (PyArg_ParseTuple(tokens, "O", &in_1)) {
+        long n_dims = longFromTwoTokens(
+          GET_, (z ? zVAR_NUMDIMS_ : rVARs_NUMDIMS_));
+        long *dims = NULL;
+        long type = (z ? typeHelper_zVAR_(NULL) : typeHelper_rVAR_(NULL));
+        void **conv_1 = allocateHyperDataStorage(z, &dims, &n_dims, type);
+        if (dims != NULL) {
+            hyperDataFromOwnedPythonSequenceTrees(
+              conv_1, in_1, dims, n_dims, type);
+            if (conv_1 != NULL) {
+                if (check(CDFlib(one, two, conv_1, NULL_))) {
+                    cleanupMultiDimensionalArray(conv_1, dims, n_dims);
+                    free(dims);
+                    return Py_None;
+                }
+                cleanupMultiDimensionalArray(conv_1, dims, n_dims);
+            }
+            free(dims);
+        }
+    }
     return NULL;
 }
 
+PyObject *tokenCustom_x_rVARs(long one, long two, PyObject *tokens,
+    long (*helper)(PyObject *)) {
+    return getHyperData(0, one, two);
+}
+
+PyObject *tokenCustom_x_zVARs(long one, long two, PyObject *tokens,
+    long (*helper)(PyObject *)) {
+    return getHyperData(1, one, two);
+}
+
+PyObject *tokenCustom_rVARs_x(long one, long two, PyObject *tokens,
+    long (*helper)(PyObject *)) {
+    return setHyperData(0, one, two, tokens);
+}
+
+PyObject *tokenCustom_zVARs_x(long one, long two, PyObject *tokens,
+    long (*helper)(PyObject *)) {
+    return setHyperData(1, one, two, tokens);
+}
+
 PyObject *tokenFormat_x_L(long one, long two, PyObject *tokens,
     long (*helper)(PyObject *)) {
     if (helper != NULL) {
@@ -1403,7 +1519,7 @@ PyObject *CdfFirstTierTokenHandler(
                         /* Borrow a reference from the tuple. */
                         PyObject *item = PySequence_GetItem(
                             output, retCounter);
-                        /* Up the refcount so we now own it. */
+                        /* Up the refn_dims so we now own it. */
                         Py_IncRef(item);
                         /* Pass ownership to the list. */
                         PyList_Append(list, item);
@@ -1540,7 +1656,11 @@ allocatedArrayFromOwnedPythonSequence(PyObject *sequence) {
                 for (i = 0; i < len; i++) {
                     item = PySequence_GetItem(sequence, i);
                     if (item != NULL) {
-                        array[i] = PyLong_AsLong(item);
+                        if (PyLong_Check(item)) { 
+                            array[i] = PyLong_AsLong(item);
+                        } else if (PyInt_Check(item)) {
+                            array[i] = PyInt_AsLong(item);
+                        }
                     } else {
                         free(array);
                         return NULL;
@@ -1589,92 +1709,71 @@ ownedPythonListFromArray(void *array, long len, long type) {
 void **
 arrayOfArrayPointers(long count) {
     if (count > 0) {
-        void **ret = (void **)calloc(count, sizeof(void *));
-
-        if (ret == NULL) {
-            printf("Failed to allocate memory for void pointer "
-                "array of size %ld.\n", count);
-            return NULL;
-        }
-        return ret;
+        return alloc(calloc(count, sizeof(void *)));
     }
     return NULL;
 }
 
-void *array(long count, long size) {
-    if (count > 0) {
-        void *ret = calloc(count, size);
-
-        if (ret == NULL) {
-            printf("Failed to allocate memory for array "
-                "of %ld items of %ld size.\n", count, size);
-            return NULL;
-        }
-        return ret;
+void *allocateArray(long count, long size) {
+    if ((count > 0) && (size > 0)) {
+        return alloc(calloc(count, size));
     }
     return NULL;
 }
 
 long *arrayOfLongs(long count) {
-    return (long *)array(count, sizeof(long));
+    return (long *)allocateArray(count, sizeof(long));
 }
 
 void **
-multiDimensionalArray(long *dims, long count, long size) {
+multiDimensionalArray(long *dims, long n_dims, long size) {
     long i;
 
-    if ((dims != NULL) && (dims[0] > 0) && (count > 0)) {
-        if (count > 1) {
-            void **level = arrayOfArrayPointers(dims[0]);
-            if (level == NULL) {
-                printf("Failed to allocate memory for array dimension.\n");
-                return NULL;
-            }
+    void **ret = NULL;
+    if (n_dims > 1) {
+        ret = arrayOfArrayPointers(dims[0]);
+        if (ret != NULL) {
             for (i = 0; i < dims[0]; i++) {
-                level[i] = multiDimensionalArray(
-                  (long *)(&(dims[1])), (count - 1), size);
+                ret[i] = multiDimensionalArray(
+                  (long *)(&(dims[1])), (n_dims - 1), size);
             }
-            return level;
-        } else {
-            return (void **)array(dims[0], size);
         }
-    } else if (size > 0) {
-        return alloc(calloc(size, 1));
+    } else {
+        ret = (void **)allocateArray((n_dims > 0 ? dims[0] : 1), size);
     }
-    return NULL;
+    return ret;
 }
 
 void
-cleanupMultiDimensionalArray(void **array, long *dims, long count) {
+cleanupMultiDimensionalArray(void **array, long *dims, long n_dims) {
     long i;
 
-    if ((array != NULL) && (dims != NULL)) {
-        if (count > 1) {
+    if (array != NULL) {
+        if (n_dims > 1) {
             for (i = 0; i < dims[0]; i++) {
                 cleanupMultiDimensionalArray(
-                    (void **)(&(array[i])), (long *)(&(dims[1])), (count - 1));
+                  array + i * sizeof(void **),
+                  (long *)(&(dims[1])), (n_dims - 1));
             }
-            free(array);
-        } else {
-            free((long *)array);
         }
+        free(array);
     }
 }
 
 PyObject *
-ownedPythonListOfListsFromArray(void **array, long *dims, long count, long type) {
+ownedPythonListOfListsFromArray(void **array, long *dims, long n_dims, long type) {
 
     PyObject *list = NULL;
     PyObject *tmp = NULL;
     long i;
 
     if ((array != NULL) && (dims != NULL)) {
-        if (count > 1) {
+        if (n_dims > 1) {
             list = PyList_New(dims[0]);
             if (list != NULL) {
                 for (i = 0; i < dims[0]; i++) {
                     tmp = ownedPythonListOfListsFromArray(
-                        (void *)(&(array[i])), (long *)(&(dims[1])), (count - 1), type);
+                        (void *)(&(array[i])), (long *)(&(dims[1])), (n_dims - 1), type);
                     if (tmp != NULL) {
                         PyList_SetItem(list, i, tmp);
                     } else {
